import sys

from pyjano.server import create_server


if __name__ == '__main__':

    server = create_server(debug=True)
    if len(sys.argv) > 1:
        server.run(port=int(sys.argv[1]))
    else:
        server.run()



"""Pyjano stands for Python Jana Orchestrator

The packet contains tools and python wrappers that allows to run JANA2 and eic-JANA frameworks
processes from python. The python also provides IPython widgets and interactive controls
"""
import itertools
import os
from time import sleep

# import click
#
# from ipywidgets import Button, IntProgress, HBox


# def is_notebook():
#     try:
#         try:
#             from IPython import get_ipython
#         except ImportError:
#             return False            # Not even IPython is installed
#
#         shell = get_ipython().__class__.__name__
#         if shell == 'ZMQInteractiveShell':
#             return True  # Jupyter notebook or qtconsole
#         elif shell == 'TerminalInteractiveShell':
#             return False  # Terminal running IPython
#         else:
#             return False  # Other type (?)
#     except NameError:
#         return False  # Probably standard Python interpreter



#
# @click.group(invoke_without_command=True)
# @click.option('--debug/--no-debug', default=False)
# @click.option('--top-dir', default="")
# @click.pass_context
# def ejpm_cli(ctx, debug, top_dir):
#     """EJPM stands for EIC Jana Packet Manager"""
#     pass



if __name__ == "__main__":
    print(is_notebook())




from flask_wtf import Form
from wtforms.fields import StringField, SubmitField
from wtforms.validators import Required


class LoginForm(Form):
    """Accepts a nickname and a room."""
    name = StringField('Name', validators=[Required()])
    room = StringField('Room', validators=[Required()])
    submit = SubmitField('Enter Chatroom')



"""Pyjano stands for Python Jana Orchestrator

Jana(nthreads=10, )
"""
import os


import escrun
from escrun.console_run_sink import ConsoleRunSink



#
# class ConsoleRunSink:
#     def add_line(self, line):
#         print(line)
#
#     def display(self):
#         print("Rendering in console")
#
#     def done(self):
#         pass
#
#     def show_running_command(self, command):
#         print(f'Command = "{command}"')
#
#     @property
#     def is_displayed(self):
#         return True
#
#
# from IPython.core.display import display
# from ipywidgets import widgets
#
#
# class NotebookRunSink:
#
#     def __init__(self):
#         self._output_widget = widgets.Output()
#         self._is_displayed = True
#         self._label = widgets.Label(value="Initializing...")
#         self._stop_button = widgets.Button(
#             description='Terminate',
#             disabled=False,
#             button_style='',  # 'success', 'info', 'warning', 'danger' or ''
#             tooltip='Terminate jana process',
#             # icon='check'
#         )
#         self._command_label = widgets.HTML()
#
#     # noinspection PyTypeChecker
#     def display(self):
#         # title_widget = widgets.HTML('<em>Vertical Box Example</em>')
#         self._stop_button.layout.display = ''
#         control_box = widgets.HBox([self._stop_button, self._label])
#
#         accordion = widgets.Accordion(children=[self._output_widget, self._command_label], selected_index=None)
#         accordion.set_title(0, 'Full log')
#         accordion.set_title(1, 'Run command')
#
#         vbox = widgets.VBox([control_box, accordion])
#
#         # display(accordion)
#         display(vbox)
#         self._is_displayed = True
#
#     def add_line(self, line):
#         to_show = [
#             'Initializing plugin',
#             'Start processing',
#             'Completed events',
#             'ERROR',
#             'FATAL',
#             '[INFO]'
#         ]
#
#         tokens = line.split('\n')
#         for token in tokens:
#             for test in to_show:
#                 if test in token:
#                     self._label.value = token
#
#         self._output_widget.append_stdout(line + '\n')
#
#     def done(self):
#         self._stop_button.layout.display = 'none'
#
#     def show_running_command(self, command):
#         tokens = shlex.split(command)
#         self._command_label.value = '<br>'.join(tokens)
#
#     @property
#     def is_displayed(self):
#         return self._is_displayed
#
#
# def _run(command, sink):
#     """Wrapper around subprocess.Popen that returns:
#
#     :return retval, start_time, end_time, lines
#     """
#     if isinstance(command, str):
#         command = shlex.split(command)
#
#     # Pretty header for the command
#     sink.add_line('=' * 20)
#     sink.add_line("RUN: " + " ".join(command))
#     sink.add_line('=' * 20)
#
#     # Record the start time
#     start_time = datetime.now()
#     lines = []
#
#     # stderr is redirected to STDOUT because otherwise it needs special handling
#     # we don't need it and we don't care as C++ warnings generate too much stderr
#     # which makes it pretty much like stdout
#     process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
#     while True:
#         line = process.stdout.readline().decode('latin-1').replace('\r', '\n')
#
#         if process.poll() is not None and line == '':
#             break
#         if line:
#             if line.endswith('\n'):
#                 line = line[:-1]
#             sink.add_line(line)
#             lines.append(line)
#
#     # Get return value and finishing time
#     retval = process.poll()
#     end_time = datetime.now()
#     sink.done()
#
#     sink.add_line("------------------------------------------")
#     sink.add_line(f"RUN DONE. RETVAL: {retval} \n\n")
#     if retval != 0:
#         sink.add_line(f"ERROR. Retval is not 0. Plese, look at the logs\n")
#
#     return retval, start_time, end_time, lines
from escrun.runner import run
from escrun.test_env import is_notebook
from .plugin import PluginFromSource, Plugin


class Jana(object):
    """Jana - class allows to configure and run ejana (and genue JANA) processes.
    It also provides IPython widgets for interactive control over jana from Jupyter
    """

    static_content_is_inserted = False

    server = None

    def __init__(self, gui='auto', **kwargs):
        self.config = {}

        if not gui:
            self.is_notebook = False
            self.sink = ConsoleRunSink()
        else:
            self.is_notebook = is_notebook()
            if (gui == 'auto' and self.is_notebook) or gui == 'notebook':
                from escrun.notebook_run_sink import NotebookRunSink
                self.sink = NotebookRunSink()
            else:
                self.sink = ConsoleRunSink()

        self.sink.to_show = ["%]", "Error", "ERROR", "FATAL"]  # "[" - Cmake like "[9%]"
        self.runner = None

        self.exec_path = 'ejana'
        self.plugin_search_paths = []
        self.config['params'] = {}
        self.config['params'].update(kwargs)

        self.config['plugins'] = {}
        self.config['flags'] = []
        self.config['input_files'] = []

        self._environ_is_updated = False

    def update_environment(self):
        """Updates the environment according to the configuration"""
        def print_plugin_locations(locations):
            for loc in locations:
                self.sink.add_line(f"   {loc}")

        ex_plugin_locations = [loc for loc in os.environ.get('JANA_PLUGIN_PATH', '').split(':') if loc]
        if ex_plugin_locations and ex_plugin_locations[0]:
            self.sink.add_line("Existing plugin locations")
            print_plugin_locations(ex_plugin_locations)

        if self.plugin_search_paths:
            self.sink.add_line("Appending them by plugin search locations")
            print_plugin_locations(self.plugin_search_paths)

        # remove a location from existing locations. We will prepend the list anyway

        env_plugin_path = ':'.join(self.plugin_search_paths + ex_plugin_locations)
        os.environ['JANA_PLUGIN_PATH'] = env_plugin_path
        # print(os.environ['JANA_PLUGIN_PATH'])
        self._environ_is_updated = True

    def configure_plugin_paths(self, plugin_paths):
        """Set additional paths where JANA looks for plugins"""
        # The later plugin location will have greater load priority
        # This means that it must be earlier in the list
        for path in plugin_paths:
            try:
                self.plugin_search_paths.remove(path)
            except ValueError:
                pass  # no such path

            self.plugin_search_paths.insert(0, path)

        self._environ_is_updated = False



    def plugin(self, plugin, **plugin_args):
        """
        Adds (activates) a plugin with a given name.
        This function may stack: .plugin('a').plugin('b')

        Example: jana.plugin('eic_smear', verbose=2, detector='beast')
                     .plugin('eventless_writer')

        :param plugin: Name of the plugin to activate
        :param plugin_args: Plugin arguments
        :return:
        """

        if plugin == 'jana' and plugin_args:
            self.config['params'].update(plugin_args)
            return self

        # It just a plugin name like:
        # .plugin('vmeson', ...)
        if isinstance(plugin, str):
            self.config['plugins'][plugin] = Plugin(plugin, **plugin_args)
            return self

        # Making sure that plugin is from Plugin
        assert isinstance(plugin, Plugin)

        # Update args
        self.config['plugins'][plugin.name] = plugin
        plugin.args.update(plugin_args)

        # Do we need to append search path?
        if plugin.search_path and plugin.search_path not in self.plugin_search_paths:
            self.plugin_search_paths.append(plugin.search_path)
            self._environ_is_updated = False    # this will trigger regenerating env

        # replace sink if needed
        if isinstance(plugin, PluginFromSource):
            plugin.builder.sink = self.sink

        return self

    def source(self, source_strings):
        """Input source (usually files) configuration.

        This function stacks with plugins and another source

        ```python
        jana.plugin('beagle_reader') \
            .source('file1.txt')     \
            .source(['file2.txt', 'file3.txt'])
        ```
        """

        if isinstance(source_strings, str):
            self.config['input_files'].append(source_strings)
        else:
            self.config['input_files'].extend(source_strings)

        return self

    def reset_plugins(self):
        """
        Resets the plugin configuration
        """

        self.config['plugins'] = {}
        return self

    def configure_plugins(self, plugins):
        """ Configures plugins (using dicts). Replaces the previous configuration.

        :param plugins:
        :return: dict with plugins config
        """
        self.reset_plugins()

        if not plugins:
            return self.config['plugins']

        def _update(plugin_name, plugin_data):
            if plugin_name == 'jana' and plugin_data:
                self.config['params'] = plugin_data
            else:
                self.config['plugins'][plugin_name] = plugin_data

        assert isinstance(plugins, list)
        for item in plugins:
            if isinstance(item, str):
                _update(item, {})
            elif isinstance(item, tuple) or isinstance(item, list):
                # plugin in form:
                # ('name', {config})
                assert isinstance(item[1], dict)
                _update(item[0], item[1])
            else:
                assert isinstance(item, dict)
                # dict must be with one key like:
                # {'beagle_reader': { ... configs ... }}
                plugin_name = list(item)[0]
                _update(plugin_name, item[plugin_name])

        return self.config['plugins']

    def configure(self, plugins=None, flags=None, files=None, params=None, plugin_paths=None):
        """Universal configuration function

        (!) this function resets and overwrites the configuration made by functions like plugin, input, etc

        Example:

        jana.configure(
        plugins=[  # a list of plugins to use:
            'beagle_reader',  # plugin name, no additional parameters
            {'open_charm': {  # add vmeson plugin & set '-Pvmeson:verbose=2' parameter
                'verbose': 1,  # Set verbose mode for that plugin
                'smearing': 1}  # Set smearing mode
            },
            {'eic_smear': {
                    'verbose': 1,
                    'detector': 'jleic'
                }
            }
        ],
        files=["/home/romanov/ceic/data/herwig6_e-p_5x100.hepmc"],  # or [list, of, files]
        params={'nthreads': 4, 'nevents': 2000}  # for parameters that don't follow <plugin>:<name> naming
                                                 # Smart enough to run it like --nthreads=8
    )  # instead of -P...


        :param plugins: list of plugins
        :param flags: list of raw flags (will be added to launch arguments without formatting)
        :param files: list of input files
        :param params: list of parameters (will be added with -P<name>=<value> flag)
        :param plugin_paths: Additional paths to search for plugins
        """
        if plugins:
            self.configure_plugins(plugins)
        if params:
            if self.config['params']:
                self.config['params'].update(params)
            else:
                self.config['params'] = params
        if flags:
            self.config['flags'] = flags
        if files:
            self.config['input_files'] = files

        if plugin_paths:
            self.configure_plugin_paths(plugin_paths)

        # noinspection PyTypeChecker
        if self.is_notebook:
            # display(Javascript('console.log("hello world")', lib='https://code.jquery.com/jquery-3.4.1.slim.js'))
            clear_output()
            # noinspection PyTypeChecker
            display(HTML('<b>JANA</b> configured...'))

    # noinspection PyTypeChecker
    def interactive_notebook(self):
        """Installs scripts and styles for page that enables interactive widgets

          Obsoletes within Jupyter lab ^1.0.2
         """

        '/static/css/bootstrap.min.css'

        display(HTML("""        
        <link rel="stylesheet" href="/static/css/bootstrap.min.css" crossorigin="anonymous">
        <script src="/static/js/bootstrap.min.js"></script>        
        <script src="/static/js/plugins.js"></script>        
        <link rel="stylesheet" href="/static/css/pretty-checkbox.min.css">
        '<b>Pyjano</b> jupyter notebook interactive loaded...'        
        """))

    # noinspection PyTypeChecker
    def plugins_gui(self):
        """Shows interactive GUI, which helps configure plugins and their parameters"""

        self.interactive_notebook()

        from pyjano.server.jana import offline_render
        display(widgets.HTML(offline_render()))
        display(Javascript("""
            if(typeof jQuery=='undefined') {
                var headTag = document.getElementsByTagName("head")[0];
                var jqTag = document.createElement('script');
                jqTag.type = 'text/javascript';
                jqTag.src = '/static/jsroot/libs/jquery.js';
                jqTag.onload = activatePluginsGui;
                headTag.appendChild(jqTag);
            } else {
                 activatePluginsGui();
            }
                """))

    def run(self, retval_raise=False):
        """Runs ejana/JANA process """
        if not self._environ_is_updated:
            self.update_environment()

        # if not self.sink.is_displayed:
        self.sink.display()

        # Build plugins?
        for plugin in self.config['plugins'].values():
            if isinstance(plugin, PluginFromSource):
                do_str = "CMake configuration"
                try:
                    plugin.builder.cmake_configure(retval_raise=True)
                    do_str = "compilation"
                    plugin.builder.build(retval_raise=True)
                except RuntimeError:
                    self.sink.add_line(f"[FATAL] during plugin '{plugin.name}' {do_str}. Look at logs/output for details")
                    return

        # apply to_show for ejana
        self.sink.to_show = [
                    'Initializing plugin',
                    'Start processing',
                    'Completed events',
                    'ERROR',
                    'FATAL',
                    '[INFO] Status:',
                    '[INFO] Start',
                    '[INFO] JPluginLoader',
                ]

        command = f"""{self.exec_path} {self.get_run_command()} -Pjana:debug_plugin_loading=1 """
        self.sink.show_running_command(command)

        # RuN
        run(command, self.sink, retval_raise=retval_raise)

    def get_run_command(self):
        """Returns the command, which is used to execute jejana"""
        add_plugins_str = "-Pplugins=" + ",".join(self.config['plugins'].keys())
        plugins_params_str = " -Pnthreads=1"
        for plugin in self.config['plugins'].values():
            for name, value in plugin.args.items():
                plugins_params_str += f' -P{plugin.name}:{name}={value}'

        params_str = " ".join([f'-P{name}={value}' for name, value in self.config['params'].items()])
        files_str = " ".join([file for file in self.config['input_files']])
        flags_str = " ".join([flag for flag in self.config['flags']])
        return f'{add_plugins_str} {plugins_params_str} {params_str}  {files_str} {flags_str}'

    def _repr_html_(self):
        plugins_str = ",".join(self.config['plugins'].keys())

        result_str = f"<strong>eJana</strong> configured<br><strong>plugins:</strong> {plugins_str}"
        if self.config['input_files']:
            sources_str = "<br>".join(self.config['input_files'])
            result_str += f"<br><strong>sources:</strong><br>{sources_str}"
        return result_str


if __name__ == "__main__":
    jana = Jana()
    jana.configure(
        plugins=[  # a list of plugins to use:
            'beagle_reader',  # plugin name, no additional parameters
            {'open_charm': {  # add vmeson plugin & set '-Pvmeson:verbose=2' parameter
                'verbose': 1,  # Set verbose mode for that plugin
                'smearing': 1}  # Set smearing mode
            },
            {'eic_smear': {
                    'verbose': 1,
                    'detector': 'jleic'
                }
            }
        ],
        files=["/home/romanov/ceic/data/herwig6_e-p_5x100.hepmc"],
        # or [list, of, files]
        params={'nthreads': 4, 'nevents': 2000}  # for parameters that don't follow <plugin>:<name> naming
        # Smart enough to run it like --nthreads=8
    )  # instead of -P...

    print(jana.get_run_command())



from flask import session
from flask_socketio import emit, join_room, leave_room
from ipyjano.server.app import socketio


@socketio.on('joined', namespace='/chat')
def joined(message):
    """Sent by clients when they enter a room.
    A status message is broadcast to all people in the room."""
    room = session.get('room')
    join_room(room)
    emit('status', {'msg': session.get('name') + ' has entered the room.'}, room=room)


@socketio.on('text', namespace='/chat')
def text(message):
    """Sent by a client when the user entered a new message.
    The message is sent to all people in the room."""
    room = session.get('room')
    emit('message', {'msg': session.get('name') + ':' + message['msg']}, room=room)


@socketio.on('left', namespace='/chat')
def left(message):
    """Sent by clients when they leave a room.
    A status message is broadcast to all people in the room."""
    room = session.get('room')
    leave_room(room)
    emit('status', {'msg': session.get('name') + ' has left the room.'}, room=room)




from flask import session, redirect, url_for, render_template, request
from . import main
from .forms import LoginForm


@main.route('/', methods=['GET', 'POST'])
def index():
    """Login form to enter a room."""
    form = LoginForm()
    if form.validate_on_submit():
        session['name'] = form.name.data
        session['room'] = form.room.data
        return redirect(url_for('.chat'))
    elif request.method == 'GET':
        form.name.data = session.get('name', '')
        form.room.data = session.get('room', '')
    return render_template('root_browser.html', form=form)


@main.route('/chat')
def chat():
    """Chat room. The user's name and room must be stored in
    the session."""
    name = session.get('name', '')
    room = session.get('room', '')
    if name == '' or room == '':
        return redirect(url_for('.index'))
    return render_template('chat.html', name=name, room=room)



# A plugin generator

from .mini_plugin_generator import generate_mini_plugin

plugin_generators = {
    'mini_plugin': generate_mini_plugin
}


def generate_plugin(plugin_type, **params):
    """ Generates a directory with jana plugin.

    Known plugin types:
        mini_plugin

    Usual parameters:
        plugin_name - snake_case defined name (directory name will correspond to it)
        class_name  - CamelCase defined name (Related C++ class names will have this name)
        path        - Path to the directory with plugin, otherwise current dir is used

    """

    # check if this plugin generator is known
    if plugin_type not in plugin_generators:
        message = f"plugin_type provided to generate_plugin function is unknown. Please use one of: " +\
                    " ".join([name for name in plugin_generators.keys()])
        raise ValueError(message)

    plugin_generators[plugin_type](**params)


import shlex
import time
from IPython.core.display import display
from ipywidgets import widgets




class NotebookRunSink:

    def __init__(self):
        self._output_widget = widgets.Output()
        self._is_displayed = True
        self._label = widgets.Label(value="Initializing...")
        self._stop_button = widgets.Button(
            description='Terminate',
            disabled=False,
            button_style='',  # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Terminate jana process',
            # icon='check'
        )
        self._command_label = widgets.HTML()

        self.to_show = []
        self.last_line_add_time = 0
        self._line_buffer=""

    # noinspection PyTypeChecker
    def display(self):
        # title_widget = widgets.HTML('<em>Vertical Box Example</em>')
        self._stop_button.layout.display = ''
        self._stop_button.layout.display = 'none'
        control_box = widgets.HBox([self._stop_button, self._label])

        accordion = widgets.Accordion(children=[self._output_widget, self._command_label], selected_index=None)
        accordion.set_title(0, 'Full log')
        accordion.set_title(1, 'Run command')

        vbox = widgets.VBox([control_box, accordion])

        # display(accordion)
        display(vbox)
        self._is_displayed = True

    def _do_add_line(self, line):
        self._output_widget.append_stdout(line + '\n')
        self.last_line_add_time = time.time()

    def add_line(self, line):
        tokens = line.split('\n')
        for token in tokens:
            for test in self.to_show:
                if test in token:
                    self._label.value = token

        # if to less time since previous add_line, buffer the input
        if time.time() - self.last_line_add_time < 0.5:
            self._line_buffer += '\n' + line
        else:
            self._do_add_line(self._line_buffer + '\n' + line)
            self._line_buffer = ""

    def done(self):
        self._stop_button.layout.display = 'none'
        if self._line_buffer:
            self._do_add_line(self._line_buffer)
            self._line_buffer = ""

    def show_running_command(self, command):
        tokens = shlex.split(command)
        self._command_label.value = '<br>'.join(tokens)

    @property
    def is_displayed(self):
        return self._is_displayed


import sys
from http.server import HTTPServer, SimpleHTTPRequestHandler


class CORSRequestHandler(SimpleHTTPRequestHandler):
    def end_headers(self):
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods', 'GET')
        self.send_header('Cache-Control', 'no-store, no-cache, must-revalidate')
        return super(CORSRequestHandler, self).end_headers()

if __name__ == '__main__':

    port = 5000
    if len(sys.argv) > 1:
        port=int(sys.argv[1])

    httpd = HTTPServer(('localhost', port), CORSRequestHandler)
    httpd.serve_forever()



import os
from pprint import pprint
import tornado.web
from notebook.utils import url_path_join
from notebook.base.handlers import IPythonHandler

import os

# i am big on absolute paths, so i made it return an absolute path
# also that makes the results clearer.
def full_path(dir_):
    if dir_[0] == '~' and not os.path.exists(dir_):
        dir_ = os.path.expanduser(dir_)
    return os.path.abspath(dir_)

class HelloWorldHandler(IPythonHandler):
    def get(self):
        self.finish('Hello, world!')


class CorsStaticFileHandler(tornado.web.StaticFileHandler):

    # def get(self, **other):
    #     pprint(other)
    #     return tornado.web.StaticFileHandler.get(**other)

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, OPTIONS')

    def get_content_type(self):
        # This is required for for jsroot to use partial file loading 
        return 'text/plain'

def load_jupyter_server_extension(nb_app):
    '''
    Register a hello world handler.

    Based on https://github.com/Carreau/jupyter-book/blob/master/extensions/server_ext.py
    '''


    web_app = nb_app.web_app

    nb_app.log.debug("=" * 30)
    nb_app.log.info('Setting Pyjano nb server extension')
    nb_app.log.debug(type(web_app))
    server_root_dir = full_path(web_app.settings['server_root_dir'])
    pprint(server_root_dir)
    # pprint(vars(web_app))
    print("=" * 30)


    host_pattern = '.*$'
    #route_pattern = url_path_join(web_app.settings['base_url'], '/hello')
    #web_app.add_handlers(host_pattern, [(route_pattern, tornado.web.StaticFileHandler)])

    web_app.add_handlers(host_pattern, [(r"/rjs/(.*)", CorsStaticFileHandler, {"path": server_root_dir, "default_filename": "README.md"})])
    #web_app.add_handlers(host_pattern, [(route_pattern, tornado.web.StaticFileHandler, {"path":r"/home/romanov/ceic/static"})])


import os

from escrun.cmake import CmakeBuildManager


class PluginCmakeBuildManager(CmakeBuildManager):
    def __init__(self, plugin_path, **kwargs):

        if not plugin_path:
            raise RuntimeError("plugin_path is not provided")
        if not os.path.isdir(plugin_path):
            raise RuntimeError("plugin_path is not a directory (or the permission is denied)")

        kwargs['plugin_path'] = plugin_path

        # Plugin name
        if 'name' not in kwargs:
            kwargs['name'] = os.path.basename(os.path.abspath(plugin_path))

        default_build_prefix = os.path.join(plugin_path, 'cmake-build')
        kwargs['build_prefix'] = kwargs.get('build_prefix', default_build_prefix)
        kwargs['install_prefix'] = kwargs.get('install_prefix', plugin_path)
        kwargs['build_target'] = kwargs.get('build_target', kwargs['name']+'_plugin')

        # For plugins set it true
        kwargs['auto_create_build_dir'] = kwargs.get('auto_create_build_dir', True)

        super().__init__(**kwargs)


class Plugin(object):
    def __init__(self, name, **kwargs):

        if not name:
            raise ValueError("Plugin name is emtpy or None. Plugin name is required")
        self._name = name

        self.args = {}
        self.args.update(kwargs)

    @property
    def name(self):
        return self._name

    @property
    def search_path(self):
        return None


class PluginFromSource(Plugin):
    """Jana plugin from """
    def __init__(self, plugin_path, **kwargs):
        self.builder = PluginCmakeBuildManager(plugin_path, **kwargs)
        self.builder.config['build_target'] = 'install'
        self.builder.config['cxx_standard'] = '17'
        self.args = {}
        super().__init__(self.builder.config['name'])

    @property
    def name(self):
        return self.builder.config['name']

    @property
    def search_path(self):
        return self.builder.config['install_prefix']

# class UserPlugin(object):
#     def


import os

mini_plugin_cpp = """
#include <JANA/JEventProcessor.h>
#include <JANA/Services/JGlobalRootLock.h>
#include <TH1D.h>
#include <TFile.h>
#include <Math/Vector4D.h>
#include <ejana/MainOutputRootFile.h>
#include <MinimalistModel/McGeneratedParticle.h>
#include <fmt/core.h>


class {{class_name}}Processor: public JEventProcessor {
private:
    std::shared_ptr<ej::MainOutputRootFile> m_file;
    int m_verbose;                /// verbosity level. 0-none, 1-some, 2-all
    TDirectory* m_plugin_dir;     /// Virtual sub-folder inside root file used for this specific plugin

    /// Declare histograms here
    TH1D* h1d_pt;

public:
    explicit {{class_name}}Processor() {
    }

    void Init() override {
        auto app = GetApplication();
        
        // Get main ROOT output file 
        m_file = app->GetService<ej::MainOutputRootFile>();

        // Set parameters that can be controlled from command line
        m_verbose = 1;
        app->SetDefaultParameter("{{plugin_name}}:verbose", m_verbose, "Verbosity level 0=none, 1=some, 2=all");

        if(m_verbose) {
            fmt::print("{{class_name}}Processor::Init()\\n  {{plugin_name}}:verbose={}\\n", m_verbose);
        }

        // Setup histograms
        m_plugin_dir = m_file->mkdir("{{plugin_name}}"); // Create a subdir inside dest_file for these results
        m_plugin_dir->cd();
        h1d_pt = new TH1D("e_pt", "electron pt", 100,0,10);
    }

    void Process(const std::shared_ptr<const JEvent>& event) override {
        // This function is called every event
        if(m_verbose == 2) fmt::print("Begin of event {} \\n", event->GetEventNumber());
                
         
        // Acquire any results you need for your analysis
        auto particles = event->Get<minimodel::McGeneratedParticle>();

        // Go through particles
        for(auto& particle: particles) {

            // select electron
            if(std::abs(particle->pdg) != 11) continue;
            if(particle->charge != -1) continue;
            if(!particle->is_stable) continue; 
            
            ROOT::Math::PxPyPzMVector p(particle->px, particle->py, particle->pz, particle->m);
            h1d_pt->Fill(p.pt());

            if(m_verbose == 2) {
                fmt::print("e E = {:<15}  pt = {:<15}  theta = {:<15.1f} \\n", p.e(), p.pt(), p.theta()*180/3.1415);
            }
        }
    }

    void Finish() override {
        fmt::print("{{class_name}}Processor::Finish(). Cleanup\\n");
    }
};

extern "C" {
    void InitPlugin(JApplication *app) {
        InitJANAPlugin(app);
        app->Add(new {{class_name}}Processor);
    }
}
"""

mini_plugin_python_example = """
# Make sure pyjano is installed in your system
# do:
#     pip install --upgrade pyjano          # for conda, venv or root install
#     pip install --user --upgrade pyjano   # for user local install
#
#
from pyjano.jana import Jana, PluginFromSource

my_plugin = PluginFromSource('./', name='{{plugin_name}}')

jana = Jana(nevents=1000, output='min_ex_output.root')
jana.plugin('lund_reader') \\
    .plugin(my_plugin, verbose=1) \\
    .source('/home/romanov/Downloads/pipi.lund') \\
    .run()
"""

mini_plugin_cmake = """
cmake_minimum_required(VERSION 3.9)

set(PLUGIN_NAME "{{plugin_name}}")

# => ADD Sources here:
set(PLUGIN_SOURCES {{class_name}}.cc )

#   Most of the time you are not interested in lines below
# ==========================================================

project(${PLUGIN_NAME})

set(PLUGIN_TARGET_NAME "${PLUGIN_NAME}_plugin")
add_library(${PLUGIN_TARGET_NAME} SHARED ${PLUGIN_SOURCES})

find_package(ROOT)

# ============ FIND JANA ===============
include(FindPackageHandleStandardArgs)

#  TODO it will be replaced when JANAConfig.cmake is finalized
if (DEFINED JANA_HOME)
    set(JANA_ROOT_DIR ${JANA_HOME})
    message(STATUS "Using JANA_HOME = ${JANA_ROOT_DIR} (From CMake JANA_HOME variable)")
elseif (DEFINED ENV{JANA_HOME})
    set(JANA_ROOT_DIR $ENV{JANA_HOME})
    message(STATUS "Using JANA_HOME = ${JANA_ROOT_DIR} (From JANA_HOME environment variable)")
else()
    message(FATAL_ERROR "Missing $JANA_HOME")
endif()

set(JANA_VERSION 2)
find_path(JANA_INCLUDE_DIR NAMES "JANA/JApplication.h" PATHS ${JANA_ROOT_DIR}/include)
find_library(JANA_LIBRARY NAMES "JANA" PATHS ${JANA_ROOT_DIR}/lib)
set(JANA_LIBRARIES ${JANA_LIBRARY})
set(JANA_INCLUDE_DIRS ${JANA_ROOT_DIR}/include)
find_package_handle_standard_args(JANA FOUND_VAR JANA_FOUND VERSION_VAR JANA_VERSION REQUIRED_VARS JANA_ROOT_DIR JANA_INCLUDE_DIR JANA_LIBRARY)

# ============ FIND EJANA ===============
#  TODO it will be replaced when EJANAConfig.cmake is finalized
if (DEFINED EJANA_HOME)
    set(EJANA_ROOT_DIR ${EJANA_HOME})
    message(STATUS "Using EJANA_HOME = ${EJANA_ROOT_DIR} (From CMake EJANA_HOME variable)")
elseif (DEFINED ENV{EJANA_HOME})
    set(EJANA_ROOT_DIR $ENV{EJANA_HOME})
    message(STATUS "Using EJANA_HOME = ${EJANA_ROOT_DIR} (From EJANA_HOME environment variable)")
else()
    message(FATAL_ERROR "Missing $EJANA_HOME")
endif()

set(EJANA_VERSION 1)
find_path(EJANA_INCLUDE_DIR NAMES "ejana/EJanaRootApplication.h" PATHS ${EJANA_ROOT_DIR}/include)
find_library(EJANA_LIBRARY NAMES "ejana" PATHS ${EJANA_ROOT_DIR}/lib)
find_library(EJANA_FMT_LIBRARY NAMES "ejana_fmt" PATHS ${EJANA_ROOT_DIR}/lib)
set(EJANA_LIBRARIES ${EJANA_LIBRARY})
set(EJANA_INCLUDE_DIRS ${EJANA_ROOT_DIR}/include)

find_package_handle_standard_args(EJANA FOUND_VAR EJANA_FOUND VERSION_VAR EJANA_VERSION REQUIRED_VARS EJANA_ROOT_DIR EJANA_INCLUDE_DIR EJANA_LIBRARY EJANA_FMT_LIBRARY)

target_include_directories(${PLUGIN_TARGET_NAME} PUBLIC ${JANA_INCLUDE_DIR} ${ROOT_INCLUDE_DIRS} ${EJANA_INCLUDE_DIR})
target_link_libraries(${PLUGIN_TARGET_NAME} ${JANA_LIB} ${ROOT_LIBRARIES} ${EJANA_FMT_LIBRARY} ${EJANA_LIBRARY})
set_target_properties(${PLUGIN_TARGET_NAME} PROPERTIES PREFIX "" OUTPUT_NAME "${PLUGIN_NAME}" SUFFIX ".so")
install(TARGETS ${PLUGIN_TARGET_NAME} LIBRARY DESTINATION "./")
"""

mini_plugin_gitignore = """
# usual build dir
cmake-build-debug/
compile/
cmake-build/

# resulting plugin
{{plugin_name}}.so

# resulting files
*.root

# IntelliJ IDEs dir
.idea/
"""

mini_plugin_readme = """

# {{plugin_name}}

## Python

Look at [pyjano_example.py](pyjano_example.py) to see the code.

**pyjano** automatically setups everything and re/builds a plugin when the code is changed. 


## Manual build and run from command line:

1. Build:

```bash
cd {{plugin_name}}
mkdir build && cd build
cmake ../
make
```

2. Jana searches compiled plugins using JANA_PLUGIN_PATH environment variable or in the current directory. 
Make sure you are running ejana where resulting .so file is located or set JANA_PLUGIN_PATH

3. Run 
```bash
ejana 
-Pplugins={{plugin_name}},hepmc_reader,open_charm
-Popen_charm:verbose=0
-Popen_charm:e_beam_energy=18
-Popen_charm:ion_beam_energy=275
-Pjana:DEBUG_PLUGIN_LOADING=1
<source >.hepmc
```
"""


def generate_mini_plugin(plugin_name, class_name, path=""):
    """Generates mini plugin"""
    plugin_dir = os.path.join(path, plugin_name) if path else plugin_name
    os.makedirs(plugin_dir, exist_ok=False)

    # Cmake file lists
    with open(os.path.join(plugin_dir, r"CMakeLists.txt"), "w+") as f:
        f.write(mini_plugin_cmake
                .replace("{{plugin_name}}", plugin_name)
                .replace("{{class_name}}", class_name))

    with open(os.path.join(plugin_dir, class_name+".cc"), "w+") as f:
        f.write(mini_plugin_cpp
                .replace("{{plugin_name}}", plugin_name)
                .replace("{{class_name}}", class_name))

    with open(os.path.join(plugin_dir, "python_example.py"), "w+") as f:
        f.write(mini_plugin_python_example
                .replace("{{plugin_name}}", plugin_name)
                .replace("{{class_name}}", class_name))

    with open(os.path.join(plugin_dir, ".gitignore"), "w+") as f:
        f.write(mini_plugin_gitignore
                .replace("{{plugin_name}}", plugin_name)
                .replace("{{class_name}}", class_name))

    with open(os.path.join(plugin_dir, "README.md"), "w+") as f:
        f.write(mini_plugin_readme
                .replace("{{plugin_name}}", plugin_name)
                .replace("{{class_name}}", class_name))

    return plugin_dir





import matplotlib.pyplot as plt
import matplotlib.animation as anim
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d import proj3d
from matplotlib.patches import FancyArrowPatch
import numpy as np


# Custom class for 3D arrows
class Arrow3D(FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)

    def set_data(self, xs, ys, zs):
        self._verts3d = xs, ys, zs


# Constants
B = 1.0
q = -1.0
m = 1.0
beta = 0  # beta = v/c. beta = 0 is the classical limit
gamma = 1 / (1 - beta ** 2)

a = q * B / (m * gamma)
v_0 = [1.0, 0.0, 2.0]


# Equations of motion r(t)
def r(t):
    return np.array([v_0[1] / a * (1 - np.cos(a * t)) + v_0[0] / a * np.sin(a * t) - v_0[1] / a,
                     -v_0[0] / a * (1 - np.cos(a * t)) + v_0[1] / a * np.sin(a * t) + v_0[0] / a, v_0[2] * t])


# First time derivative of r(t)
def v(t):
    return 0.5 * np.array(
        [v_0[0] * np.cos(a * t) + v_0[1] * np.sin(a * t), -v_0[0] * np.sin(a * t) + v_0[1] * np.cos(a * t), v_0[2]])


# Second time derivative of r(t)
def acc(t):
    return 0.5 * np.array([-a * v_0[0] * np.sin(a * t) + a * v_0[1] * np.cos(a * t),
                           -a * v_0[0] * np.cos(a * t) - v_0[1] * a * np.sin(a * t), 0])


fig = plt.figure()
ax = Axes3D(fig)
t_range = np.linspace(0, 30, 200)
r_t = r(t_range)
v_t = v(t_range)
acc_t = acc(t_range)
ax.plot(r_t[0], r_t[1], r_t[2], linestyle=':', c='black')  # Plot the path

# Plot the arrows and customize the axes
B_arrow = Arrow3D([0, 0], [0, 0], [0, 60], arrowstyle="-|>", lw=2, mutation_scale=20, color="blue")
ax.add_artist(B_arrow)

B_arrow = Arrow3D([0, -1], [0, 0], [0, 0], linestyle='--', arrowstyle="-|>", lw=1, mutation_scale=20, color="blue")
ax.add_artist(B_arrow)

point = ax.scatter(r_t[0][0], r_t[1][0], r_t[2][0], s=100, c='red')

v_arrow = Arrow3D(r_t[0][0] + [0, v_t[0][0]], r_t[1][0] + [0, v_t[1][0]], r_t[2][0] + [0, v_t[2]], arrowstyle="-|>",
                  lw=2, mutation_scale=20, color="green")
ax.add_artist(v_arrow)

a_arrow = Arrow3D(r_t[0][0] + [0, acc_t[0][0]], r_t[1][0] + [0, acc_t[1][0]], r_t[2][0] + [0, acc_t[2]],
                  arrowstyle="-|>", lw=2, mutation_scale=20, color="brown")
ax.add_artist(a_arrow)

ax.legend([B_arrow, v_arrow, a_arrow], ['Magnetic Field', 'Velocity', 'Acceleration'])
ax.set_axis_off()

plt.show()

#
#
# # Animate the plot
# def animate(num):
#     point._offsets3d = ([r_t[0][num]], [r_t[1][num]], [r_t[2][num]])
#     v_arrow.set_data(r_t[0][num] + [0, v_t[0][num]], r_t[1][num] + [0, v_t[1][num]], r_t[2][num] + [0, v_t[2]])
#     a_arrow.set_data(r_t[0][num] + [0, acc_t[0][num]], r_t[1][num] + [0, acc_t[1][num]], r_t[2][num] + [0, acc_t[2]])
#     return point,
#
#
# # Save the video
# ani = anim.FuncAnimation(fig, animate, 200, interval=100)
#ani.save("magneticmotion.mp4", fps=25, dpi=150)


def is_notebook():
    try:
        try:
            from IPython import get_ipython
        except ImportError:
            return False            # Not even IPython is installed

        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter



import os

from escrun.console_run_sink import ConsoleRunSink

from .test_env import is_notebook
from .runner import run, do_execute, sink_printer, stream_subprocess, console_printer


class CmakeBuildManager(object):
    def __init__(self, **kwargs):
        self.config = {}

        self.config.update(kwargs)  # update configs

        sink = self.config.get('sink', 'auto')
        if not sink:
            self.is_notebook = False
            self.sink = ConsoleRunSink()
        else:
            self.is_notebook = is_notebook()
            if (sink == 'auto' and self.is_notebook) or sink == 'notebook':
                # NotebookRunSink requires IPython which might not be installed if one only wants to work with console
                from escrun.notebook_run_sink import NotebookRunSink
                self.sink = NotebookRunSink()
            else:
                self.sink = ConsoleRunSink()

        default_build_prefix = os.path.join(__file__, 'cmake-build')
        self.config['build_prefix'] = self.config.get('build_prefix', default_build_prefix)

        self.config['install_prefix'] = self.config.get('install_prefix', default_build_prefix)
        self.runner = None

        # Automatically create build dir?
        self.config['auto_create_build_dir'] = self.config.get('auto_create_build_dir', False)

    def run_cmake_target(self, target, suffix='', retval_raise=False):
        """Builds G4E"""
        # Generate execution file        

        command = f"cmake --build {os.path.abspath(self.config['build_prefix'])} --target {target} {suffix}"
        
        self.sink.to_show = ["%]", "Error", "ERROR", "FATAL"]   # "[" - Cmake like "[9%]"
        self.sink.show_running_command(command)
        # if not self.sink.is_displayed:
        self.sink.display()

        retval, _,_,_ = run(command, self.sink, cwd=self.config['build_prefix'], retval_raise=retval_raise)
        return retval == 0


    def build(self, threads='auto', retval_raise=False):
        self.ensure_build_dir_exist()
        if threads == 'auto':
            import multiprocessing
            threads = multiprocessing.cpu_count()
            if threads > 2:
                threads -= 1
                
        suffix = f' -- -j {threads} -w '

        self.run_cmake_target(self.config['build_target'], suffix, retval_raise=retval_raise)

    def clean(self, retval_raise=False):
        self.run_cmake_target('clean', retval_raise=retval_raise)

    def ensure_build_dir_exist(self):
        if os.path.exists(self.config['build_prefix']):
            return      # Good! Nothing to add!

        # Create a directory?
        if self.config['auto_create_build_dir']:
            os.makedirs(self.config['build_prefix'], exist_ok=True)
            return

        # Probably it is an error
        err_msg = f"Error! CMake build path {self.config['build_prefix']} does not exist.\n"\
                  f"You may create it (if you sure it is right) with the command:\n"\
                  f"  mkdir -p {os.path.abspath(self.config['build_prefix'])}"
        raise ValueError(err_msg)

    def cmake_configure(self, build_type='RelWithDebInfo', silence_warnings=True, flags="", retval_raise=False):
        """
        Runs cmake configuration

        :param flags: CMake configuration flags like '-DCMAKE...'
        :param build_type:
        :param silence_warnings:
        :return:
        """
        self.ensure_build_dir_exist()

        if self.config.get('cxx_standard', ''):
            flags += ' -DCMAKE_CXX_STANDARD={}'.format(self.config['cxx_standard'])

        if self.config.get('install_prefix', ''):
            flags += ' -DCMAKE_INSTALL_PREFIX=' + os.path.abspath(self.config['install_prefix'])

        command = f"cmake {flags} -DCMAKE_BUILD_TYPE={build_type} {os.path.abspath(self.config['plugin_path'])}"
        self.sink.to_show = [">>>", "-- Configuring done", "-- Generating done", "Error"]  # "[" - Cmake like "[9%]"
        self.sink.show_running_command(command)

        # if not self.sink.is_displayed:
        self.sink.display()
        run(command, self.sink, cwd=self.config['build_prefix'], retval_raise=retval_raise)


    def _repr_html_(self):
        result_str = f"<strong>cmake build manager</strong><br>"
        return result_str



import shlex
import subprocess
from datetime import datetime

import sys
import asyncio
import os

if sys.platform == "win32":
    # noinspection PyUnresolvedReferences
    asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())


async def _read_stream(stream, cb):
    while True:
        line = await stream.readline()
        if line:
            cb(line)
        else:
            break


async def stream_subprocess(cmd, stdout_cb, stderr_cb):
    try:
        process = await asyncio.create_subprocess_exec(
            *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )

        await asyncio.wait(
            [
                _read_stream(process.stdout, stdout_cb),
                _read_stream(process.stderr, stderr_cb),
            ]
        )
        rc = await process.wait()
        return process.pid, rc
    except OSError as e:
        # the program will hang if we let any exception propagate
        return e


def do_execute(*aws):
    """ run the given coroutines in an asyncio loop
    returns a list containing the values returned from each coroutine.
    """
    loop = asyncio.get_event_loop()
    rc = loop.run_until_complete(asyncio.gather(*aws))
    loop.close()
    return rc

def sink_printer(sink):
    def pr(*args, **kw):
        sink.add_line(args[0])

    return pr

def console_printer(label):
    def pr(*args, **kw):
        print(label, *args, **kw)

    return pr


def name_it(start=0, template="s{}"):
    """a simple generator for task names
    """
    while True:
        yield template.format(start)
        start += 1


def runners(cmds):
    """
    cmds is a list of commands to excecute as subprocesses
    each item is a list appropriate for use by subprocess.call
    """
    next_name = name_it().__next__
    for cmd in cmds:
        name = next_name()
        out = console_printer(f"{name}.stdout")
        err = console_printer(f"{name}.stderr")
        yield stream_subprocess(cmd, out, err)


if __name__ == "__main__":
    cmd = [
        "/home/romanov/eic/g4e-dev/cmake-build-debug/g4e",
        "hepmc_full.mac",
        "-o",
        "success"
    ]


    #print(execute(*runners(cmds)))
    do_execute(stream_subprocess(cmd, console_printer("stdout"), console_printer("stderr")))


def run(command, sink, cwd=None, shell=False, retval_raise=False):
    """Wrapper around subprocess.Popen that returns:

    :return retval, start_time, end_time, lines
    """
    if isinstance(command, str):
        command = shlex.split(command)

    # Pretty header for the command
    sink.add_line('=' * 20)
    sink.add_line("CWD: " + cwd if cwd else os.getcwd())
    sink.add_line("RUN: " + " ".join(command))
    sink.add_line('=' * 20)

    # Record the start time
    start_time = datetime.now()
    lines = []

    # stderr is redirected to STDOUT because otherwise it needs special handling
    # we don't need it and we don't care as C++ warnings generate too much stderr
    # which makes it pretty much like stdout
    with subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd, shell=shell) as process:
        while True:
            line = process.stdout.readline().decode('latin-1').replace('\r', '\n')

            if process.poll() is not None and line == '':
                break
            if line:
                if line.endswith('\n'):
                    line = line[:-1]
                sink.add_line(line)
                lines.append(line)

        # Get return value and finishing time
        retval = process.poll()

    end_time = datetime.now()
    sink.add_line("------------------------------------------")
    sink.add_line(f"RUN DONE. RETVAL: {retval} \n\n")
    if retval != 0:
        sink.add_line(f"ERROR. Retval is not 0. Plese, look at the logs\n")

        if retval_raise:
            raise RuntimeError("ERROR. Retval is not 0. Plese, look at the logs")

    sink.done()
    return retval, start_time, end_time, lines


from flask import Flask


def create_server(debug=False):
    """Create an application."""
    app = Flask(__name__)
    app.debug = debug
    app.config['SECRET_KEY'] = 'gjr39dkfgdfjn344_!67#'

    from pyjano.server.jana import jana_blueprint
    app.register_blueprint(jana_blueprint)
    return app




# This is prepared plugin DB for test purpuses
from markupsafe import Markup

plugins = [
    {'name': 'lund_reader', 'type': 'reader'},
    {'name': 'beagle_reader', 'type': 'reader'},
    {'name': 'hepmc_reader', 'type': 'reader'},
    {'name': 'g4e_reader', 'type': 'reader'},
    {'name': 'gemc_reader', 'type': 'reader'},
    {'name': 'jana', 'type': ''},
    {'name': 'eic_smear', 'type': ''},
    {'name': 'event_writer', 'type': ''},
    {'name': 'vmeson', 'type': ''},
    {'name': 'open_charm', 'type': ''},
]

plugins_by_name = {p['name']: p for p in plugins}

def prepare_plugins():
    """Sets predefined values for plugins set"""

    # all plugins have some help and verbosity level
    for plugin in plugins:
        plugin['help'] = plugin['name'] + " help"
        plugin['config'] = [
            {'name': 'verbose', 'type': 'int', 'value': 0, 'help': Markup('verbosity level: 0 - silent, 2 - print all')}
        ]

    #
    # === jana - general flags ===
    plugins_by_name['jana']['help'] = Markup("""
          General configuration for JANA2 framework itself<br>          
          <br>
          <a href="https://jeffersonlab.github.io/JANA2/Installation.html" target='_blank'>More documentation</a>
          """)

    plugins_by_name['jana']['config'].extend(
        [
            {'name': 'nevents',  'type': 'int',   'value': 0,  'help': Markup('Number of events to process. 0 = all')},
            {'name': 'nskip',    'type': 'int',   'value': 0,  'help': Markup('Number of events to skip')},
            # {'name': 'nthreads', 'type': 'int',    'value': 1,  'help': Markup('Number of processing threads')},
            {'name': 'output',   'type': 'string', 'value': 'out.root', 'help': Markup('Output file name')},
        ]
    )

    #
    # === beagle_reader ===
    plugins_by_name['beagle_reader']['help'] = Markup("""
       Opens files from BeAGLE event generator as a data source<br>
       <strong>BeAGLE</strong> - <strong>Be</strong>nchmark <strong>eA</strong> <strong>G</strong>enerator for <strong>LE</strong>ptoproduction 
       <br>
       <a href="https://wiki.bnl.gov/eic/index.php/BeAGLE" target='_blank'>Documentation</a>

       """)

    # === lund_reader ===
    plugins_by_name['lund_reader']['help'] = Markup("""
       Opens files in LUND format. <br>
       The format has many flavours,  
       <a href="https://gemc.jlab.org/gemc/html/documentation/generator/lund.html"> original pythia6 </a> or 
       <a href="https://wiki.bnl.gov/eic/index.php/PYTHIA">BNL pythia 6</a> are examples      
       """)

    plugins_by_name['g4e_reader']['help'] = Markup("""
           Opens files from Geant4EIC (G4E) root flattened format.
           <a href="https://gitlab.com/jlab-eic/g4e/">G4E Official GitLab page </a>
           """)

    plugins_by_name['gemc_reader']['help'] = Markup("""
               <strong>(!) deprecated </strong> and not guaranteed to work. 
               <a href="https://gemc.jlab.org/gemc/html/index.html">GEMC page</a>
               """)

    # === EIC-smear ===
    plugins_by_name['eic_smear']['help'] = Markup("""
               Enables smearing based on EIC-Smear and JLEIC smear procedures<br>
               parameter 'detector' must be set to enable smearing               
               """)

    plugins_by_name['eic_smear']['config'].append(
        {'name': 'detector', 'type': 'string', 'value': 'jleic', 'help': Markup('Detector: beast, jleic, zeus, ephoenix ')}
    )

    # === Event writer ===
    plugins_by_name['event_writer']['help'] = Markup("""
                   Create TTree with flattened Event data. (Create branches according to data available)
                   """)

    # === Open charm analysis ===
    plugins_by_name['open_charm']['help'] = Markup("""
           Makes analysis on charm particles. Extracting basic invariant masses and other parameters with or without smearing
           """)

    plugins_by_name['open_charm']['config'].extend(
        [
            {'name': 'e_beam_energy', 'type': 'float', 'value': 5, 'help': Markup('electron beam energy GeV')},
            {'name': 'ion_beam_energy', 'type': 'float', 'value': 100, 'help': Markup('ion beam energy GeV ')},
        ]
    )

    plugins_by_name['vmeson']['config'].extend(
        [
            {'name': 'e_beam_energy', 'type': 'float', 'value': 5, 'help': Markup('electron beam energy GeV')},
            {'name': 'ion_beam_energy', 'type': 'float', 'value': 100, 'help': Markup('ion beam energy GeV ')},
        ]
    )

    return plugins


class ConsoleRunSink:

    def __init__(self):
        self.to_show = []

    def add_line(self, line):
        print(line)

    def display(self):
        print("Rendering in console")

    def done(self):
        pass

    def show_running_command(self, command):
        print(f'Command = "{command}"')

    @property
    def is_displayed(self):
        return True


import inspect
import os

import jinja2
from flask import session, redirect, url_for, render_template, request
from flask import Blueprint
from flask import Markup

from ..plugin_parser.prepared import prepare_plugins

jana_blueprint = Blueprint('main', __name__)


@jana_blueprint.route('/', methods=['GET', 'POST'])
def index():
    """Login form to enter a room."""

    return render_template('plugins.html', layout="short", plugins=prepare_plugins())


def render_jinja_html(template_loc, file_name, **context):
    return jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_loc + '/')
    ).get_template(file_name).render(context)


def offline_render():
    templates_dir = os.path.join(os.path.dirname(inspect.stack()[0][1]), 'templates')
    return render_jinja_html(templates_dir, 'plugins.html', plugins=prepare_plugins())



#
#
@jana_blueprint.route('/full', methods=['GET', 'POST'])
def full():
    """Login form to enter a room."""

    return render_template('plugins.html', layout="full", plugins=prepare_plugins(), plugin_data="""
    {'params': {'nevents': 10000, 'nthreads': 1},
     'plugins': {'beagle_reader': {}, 'event_writer': {}},
     'flags': [],
     'input_files': []}""")


@jana_blueprint.route('/start', methods=['GET', 'POST'])
def start_gui():
    """Login form to enter a room."""

    return render_template('start.html', layout="full", plugins=prepare_plugins(), plugin_data="")


@jana_blueprint.route('/chat')
def chat():
    """Chat room. The user's name and room must be stored in
    the session."""
    name = session.get('name', '')
    room = session.get('room', '')
    if name == '' or room == '':
        return redirect(url_for('.index'))
    return render_template('chat.html', name=name, room=room)


if __name__ == "__main__":
    print(offline_render())